import tempfile
from typing import Iterator, Literal, Any, TYPE_CHECKING
from pathlib import Path
import subprocess

from pydantic_compat import Field
from himena.exceptions import NotExecutable
from himena.utils.misc import PluginInfo
from himena.utils.cli import remote_to_local
from himena.workflow._base import WorkflowStep

if TYPE_CHECKING:
    from himena.types import WidgetDataModel
    from himena.workflow import Workflow


class NoParentWorkflow(WorkflowStep):
    """Describes that one has no parent."""

    output_model_type: str | None = Field(default=None)

    def iter_parents(self) -> Iterator[int]:
        yield from ()


class ProgrammaticMethod(NoParentWorkflow):
    """Describes that one was created programmatically."""

    type: Literal["programmatic"] = "programmatic"

    def _get_model_impl(self, wf):
        raise NotExecutable("Data was added programmatically, thus cannot be executed.")


class UserInput(NoParentWorkflow):
    """Describes that one will be provided as a runtime user input."""

    type: Literal["input"] = "input"
    label: str = Field(default="")
    doc: str = Field(default="")
    how: Literal["file", "model"] = "model"

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        raise NotExecutable("No user input bound to this workflow step.")


class RuntimeInputBound(UserInput):
    bound_value: Any

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        return self.bound_value


class ReaderMethod(NoParentWorkflow):
    """Describes that one was read from a file."""

    plugin: str | None = Field(default=None)
    metadata_override: Any | None = Field(default=None)

    def run(self) -> "WidgetDataModel":
        raise NotImplementedError


class LocalReaderMethod(ReaderMethod):
    """Describes that one was read from a local source file."""

    type: Literal["local-reader"] = "local-reader"
    path: Path | list[Path]

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel[Any]":
        out = self.run()
        if main := wf._mock_main_window:
            win = main.add_data_model(out)
            win._identifier = self.id
        return out

    def run(self) -> "WidgetDataModel[Any]":
        """Get model by importing the reader plugin and actually read the file(s)."""
        from himena._providers import ReaderStore
        from himena.types import WidgetDataModel

        store = ReaderStore.instance()
        model = store.run(self.path, plugin=self.plugin)
        if not isinstance(model, WidgetDataModel):
            raise ValueError(f"Expected to return a WidgetDataModel but got {model}")
        if len(model.workflow) == 0:
            model = model._with_source(
                source=self.path,
                plugin=PluginInfo.from_str(self.plugin) if self.plugin else None,
                id=self.id,
            )
        if self.metadata_override is not None:
            model.metadata = self.metadata_override
        return model


class RemoteReaderMethod(ReaderMethod):
    """Describes that one was read from a remote source file."""

    type: Literal["remote-reader"] = "remote-reader"
    host: str
    username: str
    path: Path
    port: int = Field(default=22)
    wsl: bool = Field(default=False)
    protocol: str = Field(default="rsync")
    force_directory: bool = Field(default=False)

    @classmethod
    def from_str(
        cls,
        s: str,
        /,
        wsl: bool = False,
        protocol: str = "rsync",
        output_model_type: str | None = None,
        force_directory: bool = False,
    ) -> "RemoteReaderMethod":
        username, rest = s.split("@")
        host, path = rest.split(":")
        return cls(
            username=username,
            host=host,
            path=Path(path),
            wsl=wsl,
            protocol=protocol,
            output_model_type=output_model_type,
            force_directory=force_directory,
        )

    def to_str(self) -> str:
        """Return the remote file path representation."""
        return f"{self.username}@{self.host}:{self.path.as_posix()}"

    def _get_model_impl(self, wf: "Workflow") -> "WidgetDataModel":
        out = self.run()
        if main := wf._mock_main_window:
            win = main.add_data_model(out)
            win._identifier = self.id
        return out

    def run(self):
        from himena._providers import ReaderStore

        store = ReaderStore.instance()

        with tempfile.TemporaryDirectory() as tmpdir:
            dst_path = Path(tmpdir).joinpath(self.path.name)
            self.run_command(dst_path)
            model = store.run(dst_path, plugin=self.plugin)
            model.title = self.path.name
        model.workflow = self.construct_workflow()
        return model

    def run_command(self, dst_path: Path, stdout=None):
        """Run scp/rsync command to move the file from remote to local `dst_path`."""
        args = remote_to_local(
            self.protocol,
            self.to_str(),
            dst_path,
            is_wsl=self.wsl,
            is_dir=self.force_directory,
            port=self.port,
        )
        result = subprocess.run(args, stdout=stdout)
        if result.returncode != 0:
            raise ValueError(f"Failed to run command {args}: {result!r}")



import json
from pathlib import Path
from typing import Any, TYPE_CHECKING, Callable, Literal
import warnings
from pydantic_compat import BaseModel, Field, field_validator
from himena.consts import StandardType
from himena.standards import roi
from himena.standards._base import BaseMetadata, _META_NAME

if TYPE_CHECKING:
    from pydantic import ValidationInfo

__all__ = [
    "TextMeta",
    "TableMeta",
    "DataFrameMeta",
    "DictMeta",
    "FunctionMeta",
    "DataFramePlotMeta",
    "ImageChannel",
    "DimAxis",
    "ArrayMeta",
    "ImageMeta",
    "ImageRoisMeta",
]


class TextMeta(BaseMetadata):
    """Preset for describing the metadata for a "text" type."""

    language: str | None = Field(None, description="Language of the text file.")
    spaces: int = Field(4, description="Number of spaces for indentation.")
    selection: tuple[int, int] | None = Field(None, description="Selection range.")
    font_family: str | None = Field(None, description="Font family.")
    font_size: float = Field(10, description="Font size.")
    encoding: str | None = Field(None, description="Encoding of the text file.")

    def expected_type(self):
        return StandardType.TEXT


class TableMeta(BaseMetadata):
    """Preset for describing the metadata for a "table" type."""

    current_position: list[int] | None = Field(
        default=None,
        description="Current index position of (row, column) in the table.",
    )
    selections: list[tuple[tuple[int, int], tuple[int, int]]] = Field(
        default_factory=list,
        description="Table selections in the format of ((row_start, row_end), (col_start, col_end)), where the end index is exclusive, as is always the case for the Python indexing.",
    )
    separator: str | None = Field(None, description="Separator of the table.")

    def expected_type(self):
        return StandardType.TABLE


class DataFrameMeta(TableMeta):
    """Preset for describing the metadata for a "dataframe" type."""

    transpose: bool = Field(
        default=False, description="Whether the table is transposed."
    )

    def expected_type(self):
        return StandardType.DATAFRAME


class DictMeta(BaseMetadata):
    current_tab: str | None = Field(None, description="Current tab name.")
    child_meta: dict[str, BaseMetadata] = Field(
        default_factory=dict, description="Metadata of the child models."
    )

    def expected_type(self):
        return StandardType.DICT


class FunctionMeta(BaseMetadata):
    """Preset for describing the metadata for a "function" type."""

    source_code: str | None = Field(None, description="Source code of the function.")

    def expected_type(self):
        return StandardType.FUNCTION


class DataFramePlotMeta(DataFrameMeta):
    """Preset for describing the metadata for a "dataframe.plot" type."""

    model_config = {"arbitrary_types_allowed": True}

    plot_type: Literal["line", "scatter"] = Field(
        "line", description="Type of the plot."
    )
    plot_color_cycle: list[str] | None = Field(
        None, description="Color cycle of the plot."
    )
    plot_background_color: str | None = Field(
        "#FFFFFF", description="Background color of the plot."
    )
    rois: roi.RoiListModel | Callable[[], roi.RoiListModel] = Field(
        default_factory=roi.RoiListModel, description="Regions of interest."
    )

    @classmethod
    def from_metadata(cls, dir_path: Path) -> "DataFramePlotMeta":
        self = cls.model_validate_json(dir_path.joinpath(_META_NAME).read_text())
        if (rois_path := dir_path.joinpath("rois.roi.json")).exists():
            self.rois = roi.RoiListModel.model_validate_json(rois_path.read_text())
        return self

    def unwrap_rois(self) -> roi.RoiListModel:
        """Unwrap the lazy-evaluation of the ROIs."""
        if isinstance(self.rois, roi.RoiListModel):
            return self.rois
        self.rois = self.rois()
        return self.rois

    def write_metadata(self, dir_path: Path) -> None:
        dir_path.joinpath(_META_NAME).write_text(self.model_dump_json(exclude={"rois"}))
        rois = self.unwrap_rois()
        if len(rois) > 0:
            dir_path.joinpath("rois.roi.json").write_text(
                json.dumps(rois.model_dump_typed())
            )
        return None

    def expected_type(self):
        return StandardType.DATAFRAME_PLOT


class ImageChannel(BaseModel):
    """A channel in an image file."""

    colormap: str | None = Field(None, description="Color map of the channel.")
    contrast_limits: tuple[float, float] | None = Field(
        None, description="Contrast limits of the channel."
    )
    visible: bool = Field(True, description="Whether the channel is visible.")

    @classmethod
    def default(cls) -> "ImageChannel":
        """Return a default channel (also used for mono-channel images)."""
        return cls(name=None, colormap="gray", contrast_limits=None)

    def with_colormap(self, colormap: str) -> "ImageChannel":
        """Set the colormap of the channel."""
        return self.model_copy(update={"colormap": colormap})


class DimAxis(BaseModel):
    """A dimension axis."""

    name: str = Field(..., description="Name of the axis.")
    scale: float = Field(default=1.0, description="Pixel scale of the axis.")
    origin: float = Field(default=0.0, description="Offset of the axis.")
    unit: str = Field("", description="Unit of the axis spacing.")
    labels: list[str] = Field(
        default_factory=list, description="Category labels of the axis."
    )
    default_label_format: str = Field(
        "{:s}", description="Default format of the labels."
    )

    @field_validator("name", mode="before")
    def _name_to_str(cls, v):
        return str(v)

    @classmethod
    def parse(cls, obj) -> "DimAxis":
        if isinstance(obj, str):
            axis = DimAxis(name=obj)
        elif isinstance(obj, dict):
            axis = DimAxis(**obj)
        elif isinstance(obj, DimAxis):
            axis = obj
        else:
            raise TypeError(f"Cannot convert {type(obj)} to DimAxis.")
        return axis

    def get_label(self, index: int) -> str:
        """Return the label of the axis at the given index."""
        if index < 0:
            raise ValueError("Index must be non-negative.")
        try:
            return self.labels[index]
        except IndexError:
            return self.default_label_format.format(str(index))


def __getattr__(name: str):  # pragma: no cover
    if name == "ArrayAxis":
        warnings.warn(
            "ArrayAxis is deprecated. Use DimAxis instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return DimAxis
    raise AttributeError(f"module {__name__} has no attribute {name}")


class ArrayMeta(BaseMetadata):
    """Preset for describing an array metadata."""

    axes: list[DimAxis] | None = Field(None, description="Axes of the array.")
    current_indices: tuple[int | None, ...] | None = Field(
        None, description="Current slice indices to render the array in GUI."
    )
    selections: list[tuple[tuple[int, int], tuple[int, int]]] = Field(
        default_factory=list,
        description="Selections of the array. This attribute should be any sliceable "
        "objects that can passed to the backend array object.",
    )
    unit: str | None = Field(
        None,
        description="Unit of the array values.",
    )

    def without_selections(self) -> "ArrayMeta":
        """Make a copy of the metadata without selections."""
        return self.model_copy(update={"selections": []})

    def expected_type(self):
        return StandardType.ARRAY


class ImageMeta(ArrayMeta):
    """Preset for describing an image file metadata."""

    model_config = {"arbitrary_types_allowed": True}

    channels: list[ImageChannel] = Field(
        default_factory=lambda: [ImageChannel.default()],
        description="Channels of the image. At least one channel is required.",
    )
    channel_axis: int | None = Field(None, description="Channel axis of the image.")
    is_rgb: bool = Field(False, description="Whether the image is RGB.")
    current_roi: roi.RoiModel | int | None = Field(
        None, description="Current region of interest"
    )
    current_roi_index: int | None = Field(
        None, description="Current index of the ROI in the `rois`, if applicable."
    )
    rois: roi.RoiListModel | Callable[[], roi.RoiListModel] = Field(
        default_factory=roi.RoiListModel, description="Regions of interest."
    )
    interpolation: str | None = Field(
        default=None,
        description="Interpolation method.",
    )
    skip_image_rerendering: bool = Field(
        default=False,
        description="Skip image rerendering when the model is passed to the `update_model` method. This field is only used when a function does not touch the image data itself.",
    )
    more_metadata: Any | None = Field(None, description="More metadata if exists.")

    def without_rois(self) -> "ImageMeta":
        return self.model_copy(update={"rois": roi.RoiListModel(), "current_roi": None})

    def get_one_axis(self, index: int, value: int) -> "ImageMeta":
        """Drop an axis by index for the array slicing arr[..., value, ...]."""
        if index < 0:
            index += len(self.axes)
        if index < 0 or index >= len(self.axes):
            raise IndexError(f"Invalid axis index: {index}.")
        axes = self.axes.copy()
        del axes[index]
        update = {"axes": axes, "rois": self.unwrap_rois().take_axis(index, value)}
        if (caxis := self.channel_axis) == index:
            update["channels"] = [self.channels[value]]
            update["channel_axis"] = None
            update["is_rgb"] = False
        elif caxis is not None:
            update["channel_axis"] = caxis - 1 if caxis > index else caxis
        return self.model_copy(update=update)

    def with_current_roi(self, roi: roi.RoiModel) -> "ImageMeta":
        """Set the current ROI."""
        update = {"current_roi": roi}
        if self.current_roi_index is not None:
            rois = self.unwrap_rois()
            rois.items[self.current_roi_index] = roi
            update["rois"] = rois
        return self.model_copy(update=update)

    def unwrap_rois(self) -> roi.RoiListModel:
        """Unwrap the lazy-evaluation of the ROIs."""
        if isinstance(self.rois, roi.RoiListModel):
            return self.rois
        self.rois = self.rois()
        return self.rois

    @field_validator("axes", mode="before")
    def _strings_to_axes(cls, v):
        if v is None:
            return None
        return [DimAxis.parse(axis) for axis in v]

    @field_validator("channel_axis")
    def _is_rgb_and_channels_exclusive(cls, v, values: "ValidationInfo"):
        if values.data.get("is_rgb") and v is not None:
            raise ValueError("Channel axis must be None for RGB images.")
        if v is None and len(values.data.get("channels", [])) > 1:
            raise ValueError("Channel axis is required for multi-channel images.")
        return v

    @field_validator("channels")
    def _channels_not_empty(cls, v, values: "ValidationInfo"):
        if not v:
            raise ValueError("At least one channel is required.")
        return v

    @property
    def contrast_limits(self) -> tuple[float, float] | None:
        """Return the contrast limits of the first visible channel."""
        return self.channels[0].contrast_limits

    @contrast_limits.setter
    def contrast_limits(self, value: tuple[float, float] | None):
        """Set the contrast limits of all channels."""
        for channel in self.channels:
            channel.contrast_limits = value

    @property
    def colormap(self) -> Any | None:
        """Return the colormap of the first visible channel."""
        return self.channels[0].colormap

    @colormap.setter
    def colormap(self, value: Any | None):
        """Set the colormap of all channels."""
        for channel in self.channels:
            channel.colormap = value

    @classmethod
    def from_metadata(cls, dir_path: Path) -> "ImageMeta":
        self = cls.model_validate_json(dir_path.joinpath(_META_NAME).read_text())
        if (rois_path := dir_path.joinpath(_ROIS)).exists():
            self.rois = roi.RoiListModel.model_validate_json(rois_path.read_text())
        if (cur_roi_path := dir_path.joinpath(_CURRENT_ROI)).exists():
            roi_js = json.loads(cur_roi_path.read_text())
            self.current_roi = roi.RoiModel.construct(roi_js.pop("type"), roi_js)
        if (more_meta_path := dir_path.joinpath(_MORE_META)).exists():
            with more_meta_path.open() as f:
                self.more_metadata = json.load(f)
        return self

    def write_metadata(self, dir_path: Path) -> None:
        dir_path.joinpath(_META_NAME).write_text(
            self.model_dump_json(
                exclude={"current_roi", "rois", "labels", "more_metadata"}
            )
        )
        rois = self.unwrap_rois()
        if isinstance(cur_roi := self.current_roi, roi.RoiModel):
            dir_path.joinpath(_CURRENT_ROI).write_text(
                json.dumps(cur_roi.model_dump_typed())
            )
        if len(rois) > 0:
            dir_path.joinpath(_ROIS).write_text(json.dumps(rois.model_dump_typed()))
        if (more_metadata := self.more_metadata) is not None:
            try:
                dir_path.joinpath(_MORE_META).write_text(json.dumps(more_metadata))
            except Exception as e:
                warnings.warn(
                    f"Failed to save `more_metadata`: {e}",
                    UserWarning,
                    stacklevel=2,
                )
        return None

    def expected_type(self):
        return StandardType.IMAGE


_CURRENT_ROI = "current_roi.json"
_ROIS = "rois.roi.json"
_MORE_META = "more_meta.json"


class ListMeta(BaseMetadata):
    """Preset for describing a metadata for a list-like object."""

    selections: list[int] = Field(default_factory=list)


class ImageRoisMeta(ListMeta):
    """Preset for describing an image-rois metadata."""

    axes: list[DimAxis] | None = Field(None, description="Axes of the ROIs.")

    def expected_type(self):
        return StandardType.ROIS



from __future__ import annotations

from typing import TYPE_CHECKING
from himena.types import WindowRect

if TYPE_CHECKING:
    from typing import Self


class WindowAnchor:
    def apply_anchor(
        self,
        main_window_size: tuple[int, int],
        sub_window_size: tuple[int, int],
    ) -> WindowRect | None:
        return None

    def update_for_window_rect(
        self,
        main_window_size: tuple[int, int],
        window_rect: WindowRect,
    ) -> Self:
        return self


NoAnchor = WindowAnchor()


class TopLeftConstAnchor(WindowAnchor):
    def __init__(self, left: int, top: int):
        self._left = left
        self._top = top

    def apply_anchor(
        self,
        main_window_size: tuple[int, int],
        sub_window_size: tuple[int, int],
    ) -> WindowRect:
        w, h = sub_window_size
        return WindowRect.from_tuple(self._left, self._top, w, h)

    def update_for_window_rect(
        self,
        main_window_size: tuple[int, int],
        window_rect: WindowRect,
    ) -> Self:
        return TopLeftConstAnchor(window_rect.left, window_rect.top)


class TopRightConstAnchor(WindowAnchor):
    def __init__(self, right: int, top: int):
        self._right = right
        self._top = top

    def apply_anchor(
        self,
        main_window_size: tuple[int, int],
        sub_window_size: tuple[int, int],
    ) -> WindowRect:
        main_w, _ = main_window_size
        w, h = sub_window_size
        return WindowRect.from_tuple(main_w - self._right - w, self._top, w, h)

    def update_for_window_rect(
        self,
        main_window_size: tuple[int, int],
        window_rect: WindowRect,
    ) -> Self:
        w0 = main_window_size[0]
        return TopRightConstAnchor(w0 - window_rect.right, window_rect.top)


class BottomLeftConstAnchor(WindowAnchor):
    def __init__(self, left: int, bottom: int):
        self._left = left
        self._bottom = bottom

    def apply_anchor(
        self,
        main_window_size: tuple[int, int],
        sub_window_size: tuple[int, int],
    ) -> WindowRect:
        _, main_h = main_window_size
        w, h = sub_window_size
        return WindowRect.from_tuple(self._left, main_h - self._bottom - h, w, h)

    def update_for_window_rect(
        self,
        main_window_size: tuple[int, int],
        window_rect: WindowRect,
    ) -> Self:
        h0 = main_window_size[1]
        return BottomLeftConstAnchor(window_rect.left, h0 - window_rect.bottom)


class BottomRightConstAnchor(WindowAnchor):
    def __init__(self, right: int, bottom: int):
        self._right = right
        self._bottom = bottom

    def apply_anchor(
        self,
        main_window_size: tuple[int, int],
        sub_window_size: tuple[int, int],
    ) -> WindowRect:
        main_w, main_h = main_window_size
        w, h = sub_window_size
        return WindowRect.from_tuple(
            main_w - self._right - w, main_h - self._bottom - h, w, h
        )

    def update_for_window_rect(
        self,
        main_window_size: tuple[int, int],
        window_rect: WindowRect,
    ) -> Self:
        w0, h0 = main_window_size
        return BottomRightConstAnchor(w0 - window_rect.right, h0 - window_rect.bottom)


class AllCornersAnchor(WindowAnchor):
    def apply_anchor(
        self,
        main_window_size: tuple[int, int],
        sub_window_size: tuple[int, int],
    ) -> WindowRect:
        return WindowRect.from_tuple(0, 0, *main_window_size)


def anchor_to_dict(anchor: WindowAnchor) -> dict:
    """Convert an anchor to a json-serializable dictionary."""
    if isinstance(anchor, TopLeftConstAnchor):
        return {"type": "top-left-const", "left": anchor._left, "top": anchor._top}
    if isinstance(anchor, TopRightConstAnchor):
        return {"type": "top-right-const", "right": anchor._right, "top": anchor._top}
    if isinstance(anchor, BottomLeftConstAnchor):
        return {
            "type": "bottom-left-const",
            "left": anchor._left,
            "bottom": anchor._bottom,
        }
    if isinstance(anchor, BottomRightConstAnchor):
        return {
            "type": "bottom-right-const",
            "right": anchor._right,
            "bottom": anchor._bottom,
        }
    if anchor is NoAnchor:
        return {"type": "no-anchor"}
    raise ValueError(f"Unknown anchor type: {anchor!r}")


def dict_to_anchor(data: dict) -> WindowAnchor:
    if data["type"] == "top-left-const":
        return type_to_anchor_class(data["type"])(data["left"], data["top"])
    if data["type"] == "top-right-const":
        return type_to_anchor_class(data["type"])(data["right"], data["top"])
    if data["type"] == "bottom-left-const":
        return type_to_anchor_class(data["type"])(data["left"], data["bottom"])
    if data["type"] == "bottom-right-const":
        return type_to_anchor_class(data["type"])(data["right"], data["bottom"])
    if data["type"] == "no-anchor":
        return NoAnchor
    raise ValueError(f"Unknown anchor type: {data['type']}")


def type_to_anchor_class(typ: str) -> type[WindowAnchor]:
    if typ == "top-left-const":
        return TopLeftConstAnchor
    if typ == "top-right-const":
        return TopRightConstAnchor
    if typ == "bottom-left-const":
        return BottomLeftConstAnchor
    if typ == "bottom-right-const":
        return BottomRightConstAnchor
    if typ == "no-anchor":
        return NoAnchor
    raise ValueError(f"Unknown anchor type: {typ}")



from himena.consts import MenuId
from himena.plugins import register_dock_widget_action


@register_dock_widget_action(
    title="Image Draw Tool",
    menus=[MenuId.TOOLS_DOCK],
    area="right",
    command_id="builtins:image-draw-tool",
    singleton=True,
)
def make_widget(ui):
    """Open a file explorer widget as a dock widget."""
    from himena_builtins.debug.draw_tool._widget import QImageDrawToolWidget

    return QImageDrawToolWidget(ui)
